# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Upgrade Views for ASGI
"""

import asyncio
import json
import os
from urllib.parse import parse_qs

from tailbone.views.asgi import WebsocketView
from tailbone.progress import get_basic_session


class UpgradeExecutionProgressWS(WebsocketView):

    # keep track of all "global" state for this socket
    global_state = {
        'upgrades': {},
    }

    async def __call__(self, scope, receive, send):
        app = self.get_rattail_app()

        # is user allowed to see this?
        if not await self.authorize(scope, receive, send, 'upgrades.execute'):
            return

        # keep track of client state
        client_state = {
            'uuid': app.make_uuid(),
            'disconnected': False,
            'scope': scope,
            'receive': receive,
            'send': send,
        }

        # parse upgrade uuid from query string
        query = scope['query_string'].decode('utf_8')
        query = parse_qs(query)
        uuid = query['uuid'][0]

        # first client to request progress for this upgrade, must
        # start a task to manage the collect/transmit logic for
        # progress data, on behalf of this and/or any future clients
        started_task = None
        if uuid not in self.global_state['upgrades']:

            # this upgrade is new to us; establish state and add first client
            upgrade_state = self.global_state['upgrades'][uuid] = {
                'clients': {client_state['uuid']: client_state},
            }

            # start task for transmit of progress data to all clients
            started_task = asyncio.create_task(self.manage_progress(uuid))

        else:

            # progress task is already running, just add new client
            upgrade_state = self.global_state['upgrades'][uuid]
            upgrade_state['clients'][client_state['uuid']] = client_state

        async def wait_for_disconnect():
            message = await receive()
            if message['type'] == 'websocket.disconnect':
                client_state['disconnected'] = True

        # wait forever, until client disconnects
        asyncio.create_task(wait_for_disconnect())
        while not client_state['disconnected']:

            # can stop if upgrade has completed
            if uuid not in self.global_state['upgrades']:
                break

            await asyncio.sleep(0.1)

        # remove client from global set, if upgrade still running
        if client_state['disconnected']:
            upgrade_state = self.global_state['upgrades'].get(uuid)
            if upgrade_state:
                del upgrade_state['clients'][client_state['uuid']]

        # must continue to wait for other clients, if this client was
        # the first to request progress
        if started_task:
            await started_task

    async def manage_progress(self, uuid):
        """
        Task which handles collect / transmit of progress data, for
        sake of all attached clients.
        """
        progress_session_id = 'upgrades.{}.execution_progress'.format(uuid)
        progress_session = get_basic_session(self.rattail_config,
                                             id=progress_session_id)

        upgrade_state = self.global_state['upgrades'][uuid]
        clients = upgrade_state['clients']
        while clients:

            # load latest progress data
            progress_session.load()

            # when upgrade progress is complete...
            if progress_session.get('complete'):

                # maybe set success flash msg (for all clients)
                msg = progress_session.get('success_msg')
                if msg:
                    for client in clients.values():
                        user_session = self.get_user_session(client['scope'])
                        user_session.flash(msg)
                        user_session.persist()

                # tell clients progress is complete
                for client in clients.values():
                    await client['send']({
                        'type': 'websocket.send',
                        'subtype': 'upgrades.execute_progress',
                        'text': json.dumps({'complete': True})})

                # this websocket is done, so remove all clients
                clients.clear()
                break

            # we will send this data down to client
            data = {}

            # maybe add more lines from command output
            path = self.rattail_config.upgrade_filepath(uuid, filename='stdout.log')
            offset = progress_session.get('stdout.offset', 0)
            if os.path.exists(path):
                size = os.path.getsize(path) - offset
                if size > 0:
                    with open(path, 'rb') as f:
                        f.seek(offset)
                        chunk = f.read(size)
                        data['stdout'] = chunk.decode('utf8').replace('\n', '<br />')
                    progress_session['stdout.offset'] = offset + size
                    progress_session.save()

            # send data to clients
            for client in clients.values():
                await client['send']({
                    'type': 'websocket.send',
                    'subtype': 'upgrades.execute_progress',
                    'text': json.dumps(data)})

            # pause for 1 second
            await asyncio.sleep(1)

        # no more clients, no more reason to track this upgrade
        del self.global_state['upgrades'][uuid]

    @classmethod
    def defaults(cls, config):
        cls._defaults(config)

    @classmethod
    def _defaults(cls, config):
        config.add_tailbone_websocket('upgrades.execution_progress', cls)


def defaults(config, **kwargs):
    base = globals()

    UpgradeExecutionProgressWS = kwargs.get('UpgradeExecutionProgressWS', base['UpgradeExecutionProgressWS'])
    UpgradeExecutionProgressWS.defaults(config)


def includeme(config):
    defaults(config)
